{***************************************************************************}
{ TDBFinderCombo component by Deborah Pate                                  }
{ Version 1.81, May 1999                                                    }
{                                                                           }
{ A data-aware combobox control for finding records in Delphi 2-4.          }
{ It uses the Locate method of a dataset, so works on both indexed and      }
{ unindexed fields.                                                         }
{                                                                           }
{ N.B. You can't edit records with this control - though obviously you can  }
{ edit them with other controls once you've found them with this ...        }
{ Please note that the TDBFinderCombo can only Populate itself              }
{ automatically if the Datasource it uses is created before it is - so make }
{ sure the data module is created before the form with the combo on. You    }
{ can check this by viewing the project's source and making sure you call   }
{ Application.CreateForm for the datamodule first.                          }
{                                                                           }
{***************************************************************************}
unit DBFCombo;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Forms,
  DB, DbTables, DBCtrls, DBFFuncs;

type
  TDBFinderCombo = class(TCustomComboBox)
  private
    fAutoClear: boolean;
    fBorderStyle: TBorderStyle;
    fDataLink: TFieldDataLink;
    fDisplayLink: TFieldDataLink;
    fDataList: TStringList;
    fEnterAsTab: boolean;
    fFilterMode: TFilterMode;
    fFound: boolean;
    fGroupIndex: cardinal;
    fLocateOptions: TLocateOptions;
    fOnFound: TNotifyEvent;
    fOnFoundElsewhere: TNotifyEvent;
    fOnMouseEnter: TNotifyEvent;
    fOnMouseExit: TNotifyEvent;
    fOnNotFound: TNotifyEvent;
    fOnPopulate: TNotifyEvent;
  protected
    { Field access methods }
    function GetDataSource: TDataSource;
    procedure SetDataSource(NewDataSource: TDataSource);
    function GetDataField: string;
    procedure SetDataField(const Field: string);
    function GetDisplayField: string;
    procedure SetDisplayField(const Field: string);
    function GetDisplayValue: variant;
    procedure SetBorderStyle(const Value: TBorderStyle);
    procedure SetGroupIndex(NewIdx: cardinal);
    procedure SetOnMouseEnter(const Value: TNotifyEvent);
    procedure SetOnMouseExit(const Value: TNotifyEvent);
    function GetValue: variant;
    { Other methods }
    procedure Change; override;
    procedure ClearYoungerSibs; virtual;
    procedure CMGetDataLink(var Msg: TMessage); message CM_GETDATALINK;
    procedure CMMouseEnter (var Msg: TMessage); message CM_MOUSEENTER;
    procedure CMMouseExit (var Msg: TMessage); message CM_MOUSELEAVE;
    procedure DoLocate; virtual;
    procedure DoMouseEnter; virtual;
    procedure DoMouseExit; virtual;
    procedure KeyPress(var Key: char); override;
    procedure Loaded; override;
    procedure Notification(Component: TComponent; Operation: TOperation); override;
    procedure RecordChange(Sender: TObject); virtual;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    property DataList: TStringList read fDataList write fDataList;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Populate;
    property DisplayValue: variant read GetDisplayValue;
    property Found: boolean read fFound write fFound;
    property Value: variant read GetValue;
  published
    property AutoClear: boolean read fAutoClear write fAutoClear default False;
    property BorderStyle: TBorderStyle read fBorderStyle write SetBorderStyle default bsSingle;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DataField: string read GetDataField write SetDataField;
    property DisplayField: string read GetDisplayField write SetDisplayField;
    property EnterAsTab: boolean read fEnterAsTab write fEnterAsTab default False;
    property FilterMode: TFilterMode read fFilterMode write fFilterMode default fmIgnoreFilter;
    property GroupIndex: cardinal read fGroupIndex write SetGroupIndex default 0;
    property LocateOptions: TLocateOptions read fLocateOptions write fLocateOptions default [loCaseInsensitive,loPartialKey];
    { Events }
    property OnFound: TNotifyEvent read fOnFound write fOnFound;
    property OnFoundElsewhere: TNotifyEvent read fOnFoundElsewhere write fOnFoundElsewhere;
    property OnMouseEnter: TNotifyEvent read fOnMouseEnter write SetOnMouseEnter;
    property OnMouseExit: TNotifyEvent read fOnMouseExit write SetOnMouseExit;
    property OnNotFound: TNotifyEvent read fOnNotFound write fOnNotFound;
    property OnPopulate: TNotifyEvent read fOnPopulate write fOnPopulate;
   { Inherited properties }
   {$IFDEF VER120}
      property Anchors;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ImeMode;
      property ImeName;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
   {$ENDIF}
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ItemHeight;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Style;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    { Inherited events }
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
 end;

implementation

uses
   Graphics;

{ Construction and destruction }

constructor TDBFinderCombo.Create(AOwner: TComponent);
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert(AOwner <> nil, 'TDBFinderCombo.Create can''t accept a nil parameter in its constructor');
  {$ENDIF}
  inherited Create(AOwner);

  { Set up the links to the data }
  fDataLink := TFieldDataLink.Create;
  fDataLink.Control := Self;
  fDataLink.OnDataChange := RecordChange;
  fDisplayLink := TFieldDataLink.Create;
  fDisplayLink.Control := Self;
  fDisplayLink.OnDataChange := RecordChange;

  { Initialize properties }
  fAutoClear := False;
  fBorderStyle := bsSingle;
  fEnterAsTab := False;
  fFilterMode := fmIgnoreFilter;
  fFound := False;
  fGroupIndex := 0;
  fLocateOptions := [loCaseInsensitive,loPartialKey];
  Style := csDropDownList;
  fDataList := TStringList.Create;
end;

destructor TDBFinderCombo.Destroy;
begin
  fDataLink.Free;
  fDataLink := nil;
  fDisplayLink.Free;
  fDisplayLink := nil;
  fDataList.Free;
  inherited Destroy;
end;

{ Property access methods }

function TDBFinderCombo.GetDataSource: TDataSource;
begin
  Result := fDataLink.DataSource;
end;

function TDBFinderCombo.GetDataField: string;
begin
  Result := fDataLink.FieldName;
end;

function TDBFinderCombo.GetDisplayField: string;
begin
  Result := fDisplayLink.FieldName;
end;

function TDBFinderCombo.GetDisplayValue: variant;
begin
  Result := '';
  if Assigned(fDisplayLink.Field) then
    Result := fDisplayLink.Field.Value;
end;

function TDBFinderCombo.GetValue: variant;
begin
  Result := fDataLink.Field.Value;
end;

procedure TDBFinderCombo.SetBorderStyle(const Value: TBorderStyle);
begin
  fBorderStyle := Value;
  Invalidate;
end;

procedure TDBFinderCombo.SetDataSource(NewDataSource: TDataSource);
begin
  fDataLink.DataSource := NewDataSource;
  fDisplayLink.DataSource := NewDataSource;
{ Request notification if the DataSource component is removed from the form }
  if NewDataSource <> nil then
    NewDataSource.FreeNotification(Self);
end;

procedure TDBFinderCombo.SetDataField(const Field: string);
begin
  fDataLink.FieldName := Field;
end;

procedure TDBFinderCombo.SetDisplayField(const Field: string);
begin
  if Field <> '' then begin
    fGroupIndex := 0;
  end;
  fDisplayLink.FieldName := Field;
end;

procedure TDBFinderCombo.SetGroupIndex(NewIdx: cardinal);
begin
  fGroupIndex := NewIdx;

 { Make sure that a DBFinderCombo showing a display field can't be given a
   non-zero GroupIndex }
  if ((DisplayField <> '') and (NewIdx > 0)) then
     if (MessageBox(0, PChar(SCannotGroupDisplayField), PChar(SWarning),
         MB_YESNO or MB_ICONQUESTION) = IDYES) then begin
       fDisplayLink.FieldName := '';
     end
     else begin
       fGroupIndex := 0;
     end;
end;

procedure TDBFinderCombo.SetOnMouseEnter(const Value: TNotifyEvent);
begin
  fOnMouseEnter := Value;
end;

procedure TDBFinderCombo.SetOnMouseExit(const Value: TNotifyEvent);
begin
  fOnMouseExit := Value;
end;

{****************************************************************************}
{ Other methods }

procedure TDBFinderCombo.Change;
{ Call the DoLocate method when the DBFinder's text has changed }
begin
  if (fAutoClear and (fGroupIndex > 0)) then
      ClearYoungerSibs;
  DoLocate;
  inherited Change;
  Repaint;
end;

procedure TDBFinderCombo.ClearYoungerSibs;
{ If you are searching on multiple fields and set AutoClear to True, then this
  function will be called when you change the text in a TDBFinder. It finds
  other TDBFinders with the same parent and GroupIndex, and clears any that are
  lower down the TabOrder. This is useful if you don't want your search still to
  be limited to, e.g., people with first name 'John', when you've changed the
  surname you're searching for. }
var
   i: cardinal;
   SibInfo: TDBFinderInfo;
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert(Assigned(Parent), 'TDBFinderCombo.ClearYoungerSibs: Control needs parent');
  {$ENDIF}
  for i := 0 to Parent.ControlCount - 1 do begin
     SibInfo := GetDBFinderInfo(Parent.Controls[i]);
     if (SibInfo.IsAFinder) then begin
        if ((SibInfo.TabOrder > TabOrder) and (SibInfo.GroupIndex = fGroupIndex)) then
           ClearDBFinder(Parent.Controls[i]);
     end; { if }
 end; { for }
end;

procedure TDBFinderCombo.CMGetDataLink(var Msg: TMessage);
{ Data-aware controls must respond to the Cm_GetDataLink message by returning
  a datalink reference }
begin
  Msg.Result := LongInt(fDataLink);
end;

procedure TDBFinderCombo.CMMouseEnter(var Msg: TMessage);
begin
  inherited;
  DoMouseEnter;
end;

procedure TDBFinderCombo.CMMouseExit(var Msg: TMessage);
begin
  inherited;
  DoMouseExit;
end;

procedure TDBFinderCombo.DoLocate;
{ Here's where we do the search }
var
   i: cardinal;
   FindFields: string;
   Temp, FindValues: variant;
   SibCount: cardinal;
   SibInfo: TDBFinderInfo;
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert((fDatalink.Datasource.Dataset.Active), 'TDBFinderCombo.DoLocate: table must be active');
  {$ENDIF}
   { Sort out the search parameters for a single field }
   if ((fGroupIndex = 0) or (fDisplayLink.FieldName <> '')) then begin
       if ItemIndex = -1 then begin
          Found := False;
          Exit;
       end;
       FindFields := DataField;
       if FindFields = '' then
          raise EDBFinderError.Create(SNoDatafield);
       if (fDisplayLink.FieldName = '') then
           FindValues := Text
       else begin
           FindValues := DataList[ItemIndex];
       end;
   end

   { Sort out the search parameters for multiple fields }
   else begin
      SibCount := 0;
      FindValues := VarArrayCreate([0,Parent.ControlCount],varVariant);

      { Go through the Parent's controls looking for TDBFinders }
      for i := 0 to Parent.ControlCount - 1 do begin
          SibInfo := GetDBFinderInfo(Parent.Controls[i]);

          if (SibInfo.IsAFinder) then begin
             if ((SibInfo.GroupIndex = GroupIndex) and
                ((not AutoClear) or (SibInfo.TabOrder <= TabOrder))) then begin
                { Add their Fieldnames and values to the search criteria }
                if SibCount > 0 then
                   FindFields := FindFields + ';';
                FindValues[SibCount] := SibInfo.Text;
                FindFields := FindFields + SibInfo.Datafield;
                Inc(SibCount);
             end;
          end; { if }

      end; { for }

      { Make sure at least one TDBFinder with text was found }
       if SibCount = 0 then begin
          Found := False;
          Exit;
       end;

       { Get around the bug in Locate which doesn't allow single value arrays.
         Note that this code has changed from the Delphi 2 version, which causes
         an exception in Delphi 4. }
       if SibCount = 1 then begin
          Temp := FindValues[0];
          VarClear(FindValues);
          FindValues := Temp;
       end
       else VarArrayRedim(FindValues,SibCount - 1);
   end; { GroupIndex > 0 }

   { Do the find - at last! }
   Found := fDataLink.DataSource.Dataset.Locate(FindFields,FindValues,LocateOptions);

   { Allow post-search events }
   if Found then begin
      if Assigned(FOnFound) then FOnFound(Self);
   end
   else begin
        MessageBeep(MB_ICONEXCLAMATION);
        RecordChange(Self);
        if Assigned(FOnNotFound) then FOnNotFound(Self);
   end;

end;

procedure TDBFinderCombo.DoMouseEnter;
begin
  if Assigned(fOnMouseEnter) then
    fOnMouseEnter(Self);
end;

procedure TDBFinderCombo.DoMouseExit;
begin
  if Assigned(fOnMouseExit) then
    fOnMouseExit(Self);
end;

procedure TDBFinderCombo.KeyPress(var Key: char);
{ Trap the Return key }
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert(Owner <> nil, 'TDBFinderCombo.KeyPress: control must have an owner');
  {$ENDIF}
  if Key = #13 then begin
     Key := #0;
     if EnterAsTab then
        PostMessage((Owner as TForm).Handle,WM_NEXTDLGCTL, 0, 0)
     else
        SelectAll;
  end;
   inherited KeyPress(Key);
end;

procedure TDBFinderCombo.Loaded;
{ Populate the DBFinder - can't do it in Create because the Datasource isn't ready }
begin
  inherited Loaded;
  if Assigned(fDatalink.Datasource) then
    Populate;
end;

procedure TDBFinderCombo.Notification(Component: TComponent; Operation: TOperation);
{ If the DataSource component is removed from the form, clear the
  DataSource pointer in the datalink }
begin
  inherited Notification(Component, Operation);
  if (fDataLink <> nil) and
     (Component = fDataLink.DataSource) and
     (Operation = opRemove)
  then
      fDataLink.DataSource := nil;
end;

procedure TDBFinderCombo.Populate;
{ Fill the Items property with all the possible field values }
var
   TempList: TStringList;
   CurRec: TBookmarkStr;
   TableWasOpen: boolean;
begin
  {$IFOPT D+}
    Assert((fDataLink.DataSource <> nil), 'TDBFinderCombo.Populate: Datasource is nil');
    Assert((fDataLink.DataSource.DataSet <> nil), 'TDBFinderCombo.Populate: DataSet is nil');
  {$ENDIF}
  if fDataLink.FieldName = '' then Exit;
  { Create a stringlist to help eliminate duplicate strings }
  TempList := TStringList.Create;
  try

     { Set up a temporary string list to hold the display values, so that we can
       sort it and eliminate duplicates - but *don't* do this if we need different
       data and display values, becuase it would throw them out of synch }
     if (fDisplayLink.FieldName = '') then begin
        TempList.Sorted := True;
        TempList.Duplicates := dupIgnore;
     end
     else fDataList.Clear;

     with fDatalink.Datasource.Dataset do begin
        TableWasOpen := Active;
        Open;
        if (State = dsInactive) then
          raise EDBFinderError.Create(Format(STableOpenFailed, [Name]));
        DisableControls;
        try
         { Sort out the filter mode }
          case FilterMode of
            fmApplyFilter: if Filtered = False then Filtered := True;
            fmRemoveFilter: if Filtered = True then Filtered := False;
          end;

          { Note current record }
          CurRec := Bookmark;
          First;

          { Iterate through the table, filling the stringlist(s) }
          {$IFOPT D+}
             OutputDebugString(PChar('TDBFinderCombo.Populate - iterating through table'));
          {$ENDIF}
          while not EOF do begin
              if (fDisplayLink.FieldName <> '') then begin
                 DataList.Append(VarToStr(fDataLink.Field.Value));
                 TempList.Append(VarToStr(fDisplayLink.Field.Value));
              end
              else begin
                 TempList.Append(VarToStr(fDataLink.Field.Value));
              end;
              Next;
          end;
          Bookmark := CurRec;
        finally
          EnableControls;
          Active := TableWasOpen;
        end;
     end; { with }

     { Assign the temporary stringlist's contents to the DBFinderCombo's Items }
     Items.Clear;
     Items.Assign(TempList);
     RecordChange(Self);

     { Allow post-populate events }
     if Assigned(fOnPopulate) then
        fOnPopulate(Self);
  finally
     TempList.Free;
  end;
end;

procedure TDBFinderCombo.RecordChange(Sender: TObject);
{ React to a change in the record }
begin
  if not ((fDataLink.DataSource = nil) or
    (fDataLink.DataSource.State <> dsBrowse) or
    (fDataLink.Field = nil)) then begin

    { Display the new record }
    if fDisplayLink.FieldName = '' then begin
          if (Text <> fDataLink.Field.DisplayText) then begin
              ItemIndex := Items.IndexOf(fDataLink.Field.DisplayText);
              if (Showing and (Assigned(fOnFoundElsewhere))) then
                 fOnFoundElsewhere(Self);
          end;
    end
    else if ((ItemIndex = -1) or (DataList[ItemIndex] <> fDataLink.Field.DisplayText)) then begin
         ItemIndex := DataList.IndexOf(fDataLink.Field.DisplayText);
         if (Showing and (Assigned(fOnFoundElsewhere))) then
            fOnFoundElsewhere(Self);
    end;
    Enabled := not fDataLink.Editing;
  end
  else begin

    { Record inaccessible - disable the control and clear the text if appropriate }
    if ((fDataLink.DataSource = nil) or
      (fDataLink.DataSource.State = dsInactive) or
      (fDataLink.Field = nil)) then
      ItemIndex := -1;
    Enabled := False;
  end;
end;

procedure TDBFinderCombo.WMPaint(var Message: TWMPaint);
var
  DC: HDC;
  DrawRect: TRect;
  SaveFont: HFont;
  PS: TPaintStruct;
  procedure DrawButton;
  var
    BtnState: integer;
    BtnRect: TRect;
  begin
    { Find the button coordinates }
    BtnRect := ClientRect;
    BtnRect.Left := BtnRect.Right - GetSystemMetrics(SM_CXVSCROLL);
    BtnState := DFCS_SCROLLDOWN or DFCS_FLAT;
    if not Enabled then
      BtnState := BtnState or DFCS_INACTIVE;
    if DroppedDown then
      BtnState := BtnState or DFCS_PUSHED;
    DrawFrameControl(DC, BtnRect, DFC_SCROLL, BtnState);
    DrawRect.Right := DrawRect.Right - GetSystemMetrics(SM_CXVSCROLL);
  end;
begin
  if Ctl3d then
    inherited
  else begin
    DC := Message.DC;
    if (DC = 0) then
      DC := BeginPaint(Handle, PS);
    try
      DrawRect := ClientRect;
      if (fBorderStyle = bsSingle) then
        Brush.Color := clWindowFrame;
      FrameRect(DC, DrawRect, Brush.Handle);
      InflateRect(DrawRect, -1, -1);
      Brush.Color := Color;
      FillRect(DC, DrawRect, Brush.Handle);
      DrawButton;
      DrawRect.Right := DrawRect.Right - 2;
      SaveFont := SelectObject(DC, Font.Handle);
      SetBkMode(DC, TRANSPARENT);
      SetTextColor(DC, COLORREF(Font.Color));
      ExtTextOut(DC, DrawRect.Left + 2, DrawRect.Top + 3, ETO_CLIPPED,
                 @DrawRect, PChar(Text), Length(Text), nil);
      SelectObject(DC, SaveFont);
    finally
      if Message.DC = 0 then
        EndPaint(Handle, PS);
    end;
  end;
end;



end.
